@@ -23,6 +23,9 @@ namespace PlatformQuirks {
     /** Play a system beep sound. */
     void beep();
 
+    /** Check if beep/audio notification is available on this system. */
+    bool isBeepAvailable();
+
     /** Check if the system has network connectivity. */
     bool hasNetworkConnectivity();
     
@@ -100,6 +103,18 @@ namespace PlatformQuirks {
      * @return Path to CA certificate bundle, or nullptr if not found.
      */
     const char* findCACertBundle();
+
+    /**
+     * Clear AppImage-specific environment variables before exec'ing external tools.
+     *
+     * AppImages set LD_LIBRARY_PATH and LD_PRELOAD to use bundled libraries.
+     * External tools (runuser, kde-inhibit, systemd-inhibit, etc.) need to use
+     * system libraries instead, otherwise they may fail due to symbol conflicts
+     * (e.g., PAM modules failing with "cannot open session: Module is unknown").
+     *
+     * Call this in forked child processes before execvp().
+     */
+    void clearAppImageEnvironment();
 #endif
 }
 
